#!/usr/bin/env python
# Copyright Roman Joost <roman@bromeco.de> - Rotterdam, Netherlands 2004
# this file is part of the pywifi package - a python wireless library
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import struct
import array
import math
import fcntl
import socket
import re
    
IFNAMSIZE = 16
IW_ESSID_MAX_SIZE = 16

# not used are:
# SIOCGIWNWID = 0x8B03    # get network id
# SIOCSIWCOMMIT = 0x8B00  # commiting pending changes to driver
# SIOCGIWSCAN = 0x8B19    # get scanning results

SIOCGIWNAME = 0x8B01    # get name == wireless protocol
SIOCGIWFREQ = 0x8B05    # get channel/frequency
SIOCSIWMODE = 0x8B06    # set the operation mode
SIOCGIWMODE = 0x8B07    # get operation mode
SIOCGIWSENS = 0x8B09    # get sensitivity
SIOCGIWAP = 0x8B15      # get AP MAC address
SIOCGIWRATE = 0x8B21    # get default bit rate
SIOCGIWRTS = 0x8B23     # get rts/cts threshold
SIOCGIWFRAG = 0x8B25    # get fragmention thrh
SIOCGIWTXPOW = 0x8B27   # get transmit power (dBm)
SIOCGIWRETRY = 0x8B29   # get retry limit
SIOCGIWRANGE = 0x8B0B   # range
SIOCGIWSTATS = 0x8B0F   # get wireless statistics
SIOCSIWESSID = 0x8B1A   # set essid
SIOCGIWESSID = 0x8B1B   # get essid
SIOCGIWPOWER = 0x8B2D   # get power managment settings
SIOCGIWENCODE = 0x8B2B  # get encryption information

def getNICnames():
    """ extract wireless device names of /proc/net/wireless 
        
        returns empty list if no devices are present
    """
    device = re.compile('[a-z]{3,4}[0-9]') 
    ifnames = []
    
    f = open('/proc/net/wireless', 'r')
    data = f.readlines()
    for line in data:
        try:
            ifnames.append(device.search(line).group())
        except AttributeError:
            pass 

    return ifnames

def makedict(**kwargs):
    return kwargs


class Wireless(object):
    """Access to wireless interfaces"""
    
    modes = ['Auto', 
             'Ad-Hoc',
             'Managed',
             'Master',
             'Repeat',
             'Second',
             'Monitor']
    
    def __init__(self, ifname):
        self.sockfd = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.ifname = ifname
        self.iwstruct = Iwstruct()
    
    def getAPaddr(self):
        """ returns accesspoint mac address """
        buff, s = self.iwstruct.pack_wrq(32)
        i, result = self.iwstruct.iw_get_ext(self.ifname, 
                                             SIOCGIWAP,
                                             data=s)
        if i > 0:
            return result

        return self.iwstruct.getMAC(result)
   
    def getBitrate(self):
        """ returns device bit rate """
        i, r = self.iwstruct.iw_get_ext(self.ifname, 
                                        SIOCGIWRATE)
        if i > 0:
            return r
        iwfreq = Iwfreq(r)
        return iwfreq.getBitrate()
    
    def getBitrates(self):
        """returns the number of available bitrates for the device"""
        range = Iwrange(self.ifname)
        if range.errorflag:
            return range.error
        return (range.num_bitrates, range.bitrates)

    def getChannelInfo(self):
        """returns the number of channels and available frequency for
           the device

           result is (num_of_channels, [Iwfreq objects,])
        """
        range = Iwrange(self.ifname)
        if range.errorflag:
            return range.error
        channel_info = (range.num_channels, range.frequencies)
        return channel_info

    def getEssid(self):
        """get essid information
            
           returns string
        """
        essid = ""
        buff, s = self.iwstruct.pack_wrq(32)
        i, result = self.iwstruct.iw_get_ext(self.ifname, 
                                             SIOCGIWESSID, 
                                             data=s)
        if i > 0:
            return result
        str = buff.tostring()
        return str.strip('\x00')

    def setEssid(self, essid):
        """set essid """
        raise NotImplementedError
        if len(essid) > IW_ESSID_MAX_SIZE:
            return "essid to big"
        buff, s = self.iwstruct.pack_test(essid, 32)
        i, result = self.iwstruct.iw_get_ext(self.ifname, 
                                             SIOCSIWESSID, 
                                             data=s)
        if i > 0:
            return result

    def getEncryption(self):
        """get encryption information which is probably a string of '*',
           'open', 'private'
            
           returns string
        """
        iwpoint = Iwpoint(self.ifname)
        if iwpoint.errorflag:
            return iwpoint.error
        return iwpoint.getEncryptionKey()

    def getFragmention(self):
        """returns fragmention threshold 
           
           returns int
        """
        iwparam = Iwparam(self.ifname, SIOCGIWFRAG)
        if iwparam.errorflag:
            return iwparam.error
        return iwparam.getValue()
        
    def getFrequency(self):
        """ returns frequency information """
        i, r = self.iwstruct.iw_get_ext(self.ifname, 
                                        SIOCGIWFREQ)
        if i > 0:
            return r
        iwfreq = Iwfreq(r)
        return iwfreq.getFrequency()
    
        
    def getMode(self):
        """ returns operation mode """
        i, result = self.iwstruct.iw_get_ext(self.ifname, 
                                             SIOCGIWMODE)
        if i > 0:
            return result
        mode = self.iwstruct.unpack('i', result[:4])[0]
        return self.modes[mode]

    def setMode(self, mode):
        """ sets the operation mode """
        try:
            modes = [x.lower() for x in self.modes]
            mode = mode.lower()
            wifimode = modes.index(mode)
        except ValueError:
            return "Invalid operation mode!"
        
        s = self.iwstruct.pack('I', wifimode)
        i, result = self.iwstruct.iw_get_ext(self.ifname, 
                                             SIOCSIWMODE, 
                                             data=s)
        if i > 0:
            return result
    
    def getWirelessName(self):
        """ returns wireless name """
        i, result = self.iwstruct.iw_get_ext(self.ifname, 
                                             SIOCGIWNAME)
        if i > 0:
            return result
        return result[:IFNAMSIZE]
    
    def getPowermanagment(self):
        """ returns power management settings """
        iwparam = Iwparam(self.ifname, SIOCGIWPOWER)
        if iwparam.errorflag:
            return iwparam.error
        return iwparam.getValue()

    
    def getRetrylimit(self):
        """ returns limit retry/lifetime """
        iwparam = Iwparam(self.ifname, SIOCGIWRETRY)
        if iwparam.errorflag:
            return iwparam.error
        return iwparam.getValue()
    
    def getRTS(self):
        """ returns rts """
        iwparam = Iwparam(self.ifname, SIOCGIWRTS)
        if iwparam.errorflag:
            return iwparam.error
        return iwparam.getValue()
    
    def getSensitivity(self):
        """ returns sensitivity information """
        iwparam = Iwparam(self.ifname, SIOCGIWSENS)
        if iwparam.errorflag:
            return iwparam.error
        return iwparam.getValue()
        
    def getTXPower(self):
        """ returns transmit power in dBm """
        i, r = self.iwstruct.iw_get_ext(self.ifname, 
                                        SIOCGIWTXPOW)
        if i > 0:
            return r
        iwfreq = Iwfreq(r)
        return iwfreq.getTransmitPower()
         
    def getStatistics(self):
        """returns statistics information which can also be found in
           /proc/net/wireless 
        """
        iwstats = Iwstats(self.ifname)
        return [iwstats.status, iwstats.qual, iwstats.discard,
            iwstats.missed_beacon]


class Iwstruct(object):
    """basic class to handle iwstruct data """
    
    def __init__(self):
        self.idx = 0
        self.sockfd = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def parse_data(self, fmt, data):
        """ unpacks raw C data """
        size = struct.calcsize(fmt)
        idx = self.idx

        str = data[idx:idx + size]
        self.idx = idx+size
        value = struct.unpack(fmt, str)

        # take care of a tuple like (int, )
        if len(value) == 1:
            return value[0]
        else:
            return value
    
    def pack(self, fmt, *args):
        """ calls struct.pack and returns the result """
        return struct.pack(fmt, *args)

    def pack_wrq(self, buffsize):
        """ packs wireless request data for sending it to the kernel """
        # Prepare a buffer
        # We need the address of our buffer and the size for it. The
        # ioctl itself looks for the pointer to the address in our
        # memory and the size of it.
        # Dont change the order how the structure is packed!!!
        buff = array.array('c', '\0'*buffsize)
        caddr_t, length = buff.buffer_info()
        s = struct.pack('Pi', caddr_t, length)
        return buff, s
    
    def pack_test(self, string, buffsize):
        """ packs wireless request data for sending it to the kernel """
        buffsize = buffsize - len(string)
        buff = array.array('c', string+'\0'*buffsize)
        caddr_t, length = buff.buffer_info()
        s = struct.pack('Pii', caddr_t, length, 1)
        return buff, s

    def unpack(self, fmt, packed_data):
        """ unpacks data with given format """
        return struct.unpack(fmt, packed_data)

    def _fcntl(self, request, args):
        return fcntl.ioctl(self.sockfd.fileno(), request, args)
    
    def iw_get_ext(self, ifname, request, data=None):
        """ read information from ifname """
        # put some additional data behind the interface name
        if data is not None:
            buff = IFNAMSIZE-len(ifname)
            ifreq = ifname + '\0'*buff
            ifreq = ifreq + data
        else:
            ifreq = (ifname + '\0'*32)
            
        try:
            result = self._fcntl(request, ifreq)
        except IOError, (i, e):
            return i, e
        
        return (0, result[16:])

    def getMAC(self, packed_data):
        """ extracts mac addr from packed data and returns it as str """
        mac_addr = struct.unpack('xxBBBBBB', packed_data[:8])
        return "%02X:%02X:%02X:%02X:%02X:%02X" % mac_addr

class Iwparam(object):
    """ class to hold iwparam data """

    def __init__(self, ifname, ioctl):
        # (i) value, (b) fixed, (b) disabled, (b) flags
        self.fmt = "ibbb"
        self.value = 0
        self.fixed = 0
        self.disabled = 0
        self.flag = 0
        self.errorflag = 0
        self.error = ""
        self.ioctl = ioctl 
        self.ifname = ifname
        self.update()
    
    def getValue(self):
        """ returns the value if not disabled """
        if self.disabled:
            return 'off'
        else:
            return self.value
    
    def update(self):
        iwstruct = Iwstruct()
        i, r = iwstruct.iw_get_ext(self.ifname, 
                                   self.ioctl)
        if i > 0:
            self.errorflag = i
            self.error = r
        self._parse(r)
    
    def _parse(self, data):
        """ unpacks iwparam data """
        iwstruct = Iwstruct()
        self.value, self.fixed, self.disabled, self.flag =\
            iwstruct.parse_data(self.fmt, data)
        

class Iwfreq(object):
    """ class to hold iwfreq data
        delegates to Iwstruct class
    """
    
    KILO = 10**3
    MEGA = 10**6
    GIGA = 10**9
    
    def __init__(self, data=None):
        self.fmt = "ihbb"
        if data is not None:
            self.frequency = self.parse(data)
        else:
            self.frequency = 0
        self.iwstruct = Iwstruct()
        
    def __getattr__(self, attr):
        return getattr(self.iwstruct, attr)

    def parse(self, data):
        """ unpacks iwparam"""
        
        size = struct.calcsize(self.fmt)
        m, e, i, pad = struct.unpack(self.fmt, data[:size])
        # XXX well, its not *the* frequency - we need a better name
        if e == 0:
            return m
        else:
            return float(m)*10**e
    
    def getFrequency(self):
        """returns Frequency (str) 
            
           data - binary data returned by systemcall (iw_get_ext())
        """
        freq = self.frequency
        
        if freq >= self.GIGA:
            return "%0.3fGHz" %(freq/self.GIGA)

        if freq >= self.MEGA:
            return "%0.3fMHZ" %(freq/self.MEGA)

        if freq >= self.KILO:
            return "%0.3fKHz" %(freq/self.KILO)
    
    def getBitrate(self):
        """ returns Bitrate in Mbit 
        
           data - binary data returned by systemcall (iw_get_ext())
        """
        bitrate = self.frequency

        if bitrate >= self.GIGA:
            return "%i Gb/s" %(bitrate/self.GIGA)

        if bitrate >= self.MEGA:
            return "%i Mb/s" %(bitrate/self.MEGA)
        
        if bitrate >= self.KILO:
            return "%i Kb/s" %(bitrate/self.KILO)

    def getTransmitPower(self):
        """ returns transmit power in dbm """
        # XXX something flaky is going on with m and e
        # eg. m = 50 and e should than be 0, because the number is stored in
        # m and don't needs to be recalculated
        return "%i dBm" %self.mw2dbm(self.frequency/10)
    
    def getChannel(self, freq, iwrange):
        """returns channel information given by frequency
           
           returns None if frequency can't be converted
           freq = frequency to convert (int)
           iwrange = Iwrange object
        """
        if freq < self.KILO:
            return None
        
        # XXX
        # for frequency in iwrange.frequencies

        
    def mw2dbm(self, mwatt):
        """ converts mw to dbm(float) """
        return math.ceil(10.0 * math.log10(mwatt))
        
    def _setFrequency(self, list):
        """sets self.frequency by given list 
           
           currently only used by Iwrange
        """
        assert len(list) == 4
        m, e, i, pad = list
        if e == 0:
            self.frequency = m
        else:
            self.frequency = float(m)*10**e

class Iwstats(object):
    """ class to hold iwstat data """

    def __init__(self, ifname):
        # (2B) status, 4B iw_quality, 6i iw_discarded
        self.fmt = "2B4B6i"
        self.status = 0
        self.qual = Iwquality()
        self.discard = {}
        self.missed_beacon = 0
        self.ifname = ifname
        self.update()

    def update(self):
        iwstruct = Iwstruct()
        buff, s = iwstruct.pack_wrq(32)
        i, result = iwstruct.iw_get_ext(self.ifname, 
                                        SIOCGIWSTATS, 
                                        data=s)
        self._parse(buff.tostring())
    
    def _parse(self, data):
        """ unpacks iwstruct data """
        struct = Iwstruct()
        iwqual = Iwquality()
        iwstats_data = struct.parse_data(self.fmt, data)
        
        self.status = iwstats_data[0:2]
        self.qual.quality, self.qual.sl, self.qual.nl,\
            self.qual.flags = iwstats_data[2:6]
        nwid, code, frag, retries, flags = iwstats_data[6:11]
        self.missed_beacon = iwstats_data[11:12][0]
        self.discard = makedict(nwid=nwid, code=code,
            fragment=frag, retries=retries, misc=flags)

class Iwquality(object):
    """ class to hold iwquality data """

    def __init__(self):
        self.quality = 0
        self.sl = 0
        self.nl = 0
        self.updated = 0
        self.fmt = "4B"

    def parse(self, data):
        """ unpacks iwquality data """
        struct = Iwstruct()
        qual, sl, nl, flags = struct.parse_data(self.fmt, data)

        # compute signal and noise level
        self.signal_level = sl
        self.noise_level = nl

        # asign the other values
        self.quality = qual
        self.updated = flags

    def setValues(self, list):
        """ assigns values given by a list to our attributes """
        attributes = ["quality", "signallevel", "noise_level",
            "updated"]
        assert len(list) == 4
        
        for i in range(len(list)):
            setattr(self, attributes[i], list[i])
    
    def getSignallevel(self):
        """ returns signal level """
        return self.sl-0x100

    def setSignallevel(self, sl):
        """ sets signal level """
        self.sl = sl
    signallevel = property(getSignallevel, setSignallevel)
    
    def getNoiselevel(self):
        """ returns noise level """
        return self.nl - 0x100

    def setNoiselevel(self):
        raise NotImplementedError
        self.nl = nl
    noiselevel = property(getNoiselevel, setNoiselevel)

class Iwpoint(object):
    """ class to hold iwpoint data """

    IW_ENCODE_DISABLED = 0x8000     # encoding is disabled
    IW_ENCODE_NOKEY = 0x0800      # key is write only, not present

    def __init__(self, ifname):
        self.key = [0,0,0,0]
        self.fields = 0
        self.flags = 0
        # (4B) pointer to data, H length, H flags
        self.fmt = "4BHH"
        self.errorflag = 0
        self.error = ""
        self.ifname = ifname
        self.update()

    def __getattr__(self, attr):
        return getattr(self.iwstruct, attr)
    
    def update(self):
        iwstruct = Iwstruct()
        buff, s = iwstruct.pack_wrq(32)
        i, result = iwstruct.iw_get_ext(self.ifname, 
                                        SIOCGIWENCODE, 
                                        data=s)
        if i > 0:
            self.errorflag = i
            self.error = result
        self._parse(result)
        
    def getEncryptionKey(self):
        """ returns encryption key as '**' or 'off' as str """
        if self.flags & self.IW_ENCODE_DISABLED != 0:
            return 'off'
        elif self.flags & self.IW_ENCODE_NOKEY != 0:
            # a key is set, so print it
            return '**' * self.fields
    
    def _parse(self, data):
        """ unpacks iwpoint data
        """
        iwstruct = Iwstruct()
        ptr, ptr, ptr, ptr, self.fields, self.flags =\
            iwstruct.parse_data(self.fmt, data)
        self.key = [ptr, ptr, ptr, ptr]

class Iwrange(object):
    """holds iwrange struct """
    IW_MAX_FREQUENCIES = 32

    def __init__(self, ifname):
        self.fmt = "iiihb6ii4B4Bi32i2i2i2i2i3h8h2b2bhi8i2b3h2i2ihB17x"\
            + self.IW_MAX_FREQUENCIES*"ihbb"
        
        self.ifname = ifname
        self.errorflag = 0
        self.error = ""
        
        # informative stuff
        self.throughput = 0
        
        # nwid (or domain id)
        self.min_nwid = self.max_nwid = 0
        
        # frequency for backward compatibility
        self.old_num_channels = self.old_num_frequency = self.old_freq = 0
        
        # signal level threshold
        self.sensitivity = 0
        
        # link quality
        self.max_qual = Iwquality()
        self.avg_qual = Iwquality()

        # rates
        self.num_bitrates = 0
        self.bitrates = []

        # rts threshold
        self.min_rts = self.max_rts = 0

        # fragmention threshold
        self.min_frag = self.max_frag = 0

        # power managment
        self.min_pmp = self.max_pmp = 0
        self.min_pmt = self.max_pmt = 0
        self.pmp_flags = self.pmt_flags = self.pm_capa = 0

        # encoder stuff
        self.encoding_size = 0
        self.num_encoding_sizes = self.max_encoding_tokens = 0
        self.encoding_login_index = 0

        # transmit power
        self.txpower_capa = self.num_txpower = self.txpower = 0

        # wireless extension version info
        self.we_vers_compiled = self.we_vers_src = 0

        # retry limits and lifetime
        self.retry_capa = self.retry_flags = self.r_time_flags = 0
        self.min_retry = self.max_retry = 0
        self.min_r_time = self.max_r_time = 0

        # frequency
        self.num_channels = self.num_frequency = 0
        self.frequencies = []
        self.update()
    
    def update(self):
        """updates Iwrange object by a system call to the kernel 
           and updates internal attributes
        """
        iwstruct = Iwstruct()
        buff, s = iwstruct.pack_wrq(640)
        i, result = iwstruct.iw_get_ext(self.ifname, 
                                        SIOCGIWRANGE, 
                                        data=s)
        if i > 0:
            self.errorflag = i
            self.error = result
        data = buff.tostring()
        self._parse(data)
        
    def _parse(self, data):
        struct = Iwstruct()
        result = struct.parse_data(self.fmt, data)
        
        # XXX there is maybe a much more elegant way to do this
        self.throughput, self.min_nwid, self.max_nwid = result[0:3]
        self.old_num_channels, self.old_num_frequency = result[3:5]
        self.old_freq = result[5:11]
        self.sensitivity = result[11]
        self.max_qual.setValues(result[12:16])
        self.avg_qual.setValues(result[16:20])
        self.num_bitrates = result[20] # <- XXX
        raw_bitrates = result[21:53]
        for rate in raw_bitrates:
            iwfreq = Iwfreq()
            iwfreq.frequency = rate
            self.bitrates.append(iwfreq)
            
        self.min_rts, self.max_rts = result[53:55]
        self.min_frag, self.max_frag = result[55:57]
        self.min_pmp, self.max_pmp = result[57:59]
        self.min_pmt, self.max_pmt = result[59:61]
        self.pmp_flags, self.pmt_flags, self.pm_capa = result[61:64]
        self.encoding_size = result[64:72]
        self.num_encoding_sizes, self.max_encoding_tokens = result[72:74]
        self.encoding_login_index = result[74:76]
        self.txpower_capa, self.num_txpower = result[76:78]
        self.txpower = result[78:86]
        self.we_vers_compiled, self.we_vers_src = result[86:88]
        self.retry_capa, self.retry_flags, self.r_time_flags = result[88:91]
        self.min_retry, self.max_retry = result[91:93]
        self.min_r_time, self.max_r_time = result[93:95]
        self.num_channels = result[95]
        self.num_frequency = result[96]
        freq = result[97:]
        
        i = self.num_frequency
        for x in range(0, len(freq), 4):
            iwfreq = Iwfreq()
            iwfreq._setFrequency(freq[x:x+4])
            self.frequencies.append(iwfreq)
            i -= 1
            if i <= 0:
                break
        

#$LastChangedDated$
#$Rev: 37 $
